use axum::{extract::State, routing::post, Json, Router};

use crate::auth::verify_password_hash;
use crate::database::model::member::SearchMember;
use crate::database::model::Member as DbMember;
use crate::database::model::Session as DbSession;
use crate::database::model::User as DbUser;
use crate::database::model::UserMember as DbUserMember;
use crate::model::session::Session;
use crate::{auth::generate_password_hash, AppState};

pub fn routes() -> Router<AppState> {
    Router::new()
        .route("/auth/login", post(login))
        .route("/auth/register", post(register))
}

#[derive(serde::Deserialize)]
pub struct LoginRequest {
    email: String,
    password: String,
}

pub async fn login<'a>(
    State(state): State<AppState>,
    Json(login_request): Json<LoginRequest>,
) -> Result<String, crate::Error> {
    let db_user = DbUser::get_from_email(&state.pool, login_request.email).await?;

    match verify_password_hash(&login_request.password, &db_user.password).await {
        Ok(_) => (),
        Err(_err) => return Err(crate::Error::Auth(crate::auth::AuthError::InvalidPassword)),
    };

    // Create session
    let mut transaction = state.pool.begin().await?;

    let db_session: DbSession = Session::new(db_user.user_id).into();
    db_session.insert(&mut transaction).await?;

    transaction.commit().await?;

    Ok(db_session.token)
}

#[derive(serde::Deserialize)]
pub struct RegisterRequest {
    email: String,
    password: String,
    registration_tokens: Vec<String>,
}

pub async fn register<'a>(
    State(state): State<AppState>,
    Json(auth_request): Json<RegisterRequest>,
) -> Result<String, crate::Error> {
    // Get all the members to link with the user
    let members = DbMember::search(
        &state.pool,
        SearchMember {
            registration_tokens: Some(auth_request.registration_tokens),
        },
    )
    .await?;

    let member_ids: Vec<String> = members.into_iter().map(|m| m.member_id).collect();

    // Hash password
    let password_hash = match generate_password_hash(auth_request.password).await {
        Ok(hash) => hash,
        Err(_err) => return Err(crate::Error::Auth(crate::auth::AuthError::InvalidToken)),
    };

    let mut transaction = state.pool.begin().await?;

    // Insert the user to the database
    let user_id = DbUser::insert(&mut transaction, &auth_request.email, &password_hash).await?;

    // Link the user to the members
    let user_ids: Vec<uuid::Uuid> = vec![user_id; member_ids.len()];
    DbUserMember::insert_many(&mut transaction, &user_ids, &member_ids).await?;

    // Create a new session
    let db_session: DbSession = Session::new(user_id).into();
    db_session.insert(&mut transaction).await?;

    transaction.commit().await?;

    Ok(db_session.token)
}
