use crate::auth::AuthError;
use axum::{
    http::StatusCode,
    response::{IntoResponse, Response},
    Json,
};

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("CSV error: {0}")]
    Csv(#[from] csv::Error),

    #[error("Auth error: {0}")]
    Auth(#[from] crate::auth::AuthError),

    #[error("Database error: {0}")]
    Database(#[from] sqlx::Error),

    #[error("Resource not found")]
    NotFound,
}

impl IntoResponse for Error {
    fn into_response(self) -> Response {
        tracing::error!("Error... {:?}", self);

        let (status, error_message) = match self {
            Error::Auth(AuthError::NoPermssions) => {
                (StatusCode::UNAUTHORIZED, String::from("No permissions"))
            }
            Error::Auth(AuthError::InvalidToken) => {
                (StatusCode::BAD_REQUEST, String::from("Invalid token"))
            }
            Error::Auth(AuthError::Unexpected) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                String::from("Unexpected error occured"),
            ),
            Error::Auth(AuthError::InvalidPassword) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                String::from("Invalid password"),
            ),
            Error::Csv(err) => (StatusCode::BAD_REQUEST, err.to_string()),
            Error::NotFound => (
                StatusCode::BAD_REQUEST,
                String::from("Could not find resource"),
            ),
            Error::Database(err) => (StatusCode::INTERNAL_SERVER_ERROR, err.to_string()),
        };

        let body = Json(serde_json::json!({
            "error": error_message
        }));

        (status, body).into_response()
    }
}
